/*
 * ouvej, determines what, when and where the next lesson will be
 * Copyright (C) 2003 Olivier Mehani <shtrom@ssji.net, olmehani@etu.utc.fr>
 * 
 * $Id: btree.c,v 1.9 2003/12/08 19:00:56 shtrom Exp $
 * binary tree handling functions 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <limits.h> // for INT_MIN
#include "ouvej.h"

/*
 * The magic number is (4+MAJOR)*16+2*MINOR
 */
int cachemagic=0x42;

/*
 * the binary tree will store lower values on the right (earlier) and greater
 * or equal values on the left (later)
 */
bt_t *btree_make(ll_t *list)
{
	ll_t *cur_list=list;
	bt_t *btree=NULL;

//	DPRINT("making binary tree from list at %p\n",list);

	while(cur_list) {
		btree_add(&btree, cur_list->lesson);
//		DPRINT("pointers: current=%p next=%p\n",cur_list, cur_list->next);
		cur_list=cur_list->next;
	}
	
	return btree;
}

bt_t *btree_make_equilibrated(int size, lesson_t lessons[size])
{
	bt_t *btree;
	int half=size/2;

	if(size<1) {
		DPRINT("not enough entries (%d)\n",size);
		return NULL;
	}

	DPRINT("working with %d entries\n",size);
	
	DPRINT("pivot item at index %d\n",half);

	/*
	 * add the pivot as the root for the rest 
	 */
	btree=btree_new_node(lessons[half]);

	/*
	 * build the lower part tree and link it
	 */
	btree->earlier=btree_make_equilibrated(half,lessons);
	
	/*
	 * build the upper part and link it
	 *
	 * the number of elements to consider depends on whether the
	 * total number of elements was odd or even, we don't want to
	 * forget the last element... 
	 */
	btree->later=btree_make_equilibrated(size%2?(half):(half-1),&lessons[half+1]);
	
	return btree;
}

void btree_add(bt_t **btree, lesson_t l)
{
	bt_t *cur=*btree;

	DPRINT("adding lesson %s\n", l.name);

	if(!cur) {
		DPRINT("new tree\n");
		*btree=btree_new_node(l);
	} else {
		while(cur) {
			if(lesson_compare_beg(l,cur->lesson)<0) {
				DPRINT("lesson occurs earlier than %s\n",cur->lesson.name);
				if(cur->earlier==NULL) {
					cur->earlier=btree_new_node(l);
					cur=NULL;
				} else
					cur=cur->earlier;
			} else {
				DPRINT("lesson occurs later than %s\n",cur->lesson.name);
				if(cur->later==NULL) {
					cur->later=btree_new_node(l);
					cur=NULL;
				} else
					cur=cur->later;
			}	
		}
	}
}

bt_t *btree_new_node(lesson_t l)
{
	bt_t *node;

	DPRINT("allocating new node\n");

	node=malloc(sizeof(bt_t));
	if(!node) {
		ERROR("couldn't allocate memory\n");
		exit(1);
	}

	node->lesson=l;
	node->earlier=NULL;
	node->later=NULL;
	
	return node;
}

int btree_write(bt_t *btree, char *filename)
{
	lesson_t l;
	FILE *out; 

	DPRINT("writing to %s\n",filename);

	/*
	 * ok, on for the dirty memory dump :-)
	 * this means "On y va bourrin"
	 */
	out=fopen(filename,"w");
	if(!out) {
		ERROR("can't open cache file %s for writing (%s)\n",filename,strerror(errno));
		exit(1);
	}

	/*
	 * Good Idea (TM) : the file is a memory dump, and as the data
	 * structure may be subject to changes, it's better to specify a magic
	 * number to correctly handle the cache file later
	 *
	 * I know it is a dirty way of storing data, but hey, this is a beta
	 * version I don't really have clean programing in mind right now !
	 *
	 * And if you don't like this, the next version will be coded in LISP !
	 */
	fwrite(&cachemagic,sizeof(cachemagic),1,out);

	l=btree_get_smallest_and_remove(&btree);
	while (*l.name!=0) {
		DDO(lesson_show(l));
		fwrite(&l,sizeof(lesson_t),1,out);
		l=btree_get_smallest_and_remove(&btree);
	}

	fclose(out);
	
	return 0;
}

/*
 * ReadBufferIncrement every realloc will allocate memory space for RBI lessons
 */
#define RBI	5

bt_t *btree_read(char *filename)
{
	bt_t *btree;
	int filemagic;
	lesson_t *llist=NULL;
	int num=0, remains=0;
	FILE *in; 

	DPRINT("reading btree from %s\n",filename);

	in=fopen(filename,"r");
	if(!in) {
		ERROR("can't open cache file %s for reading (%s)\n",filename,strerror(errno));
		exit(1);
	}

	fread(&filemagic,sizeof(filemagic),1,in);
	if(filemagic!=cachemagic) {
		DPRINT("magic number mismatch: expected 0x%x, got 0x%x, regenerating\n",cachemagic,filemagic);
		/*
		 * TODO: better magic number check and regenerate the cache if
		 * needed
		 */
		/*
		 * I think i worked my LISP too much...
		 */
		//cache_regen(timetable_path(get_homedir(NULL)),cache_path(get_homedir(NULL)));	
		exit(1);
	}

	DPRINT("magic number 0x%x ok\n",filemagic);

	/*
	 * Note for further modifications: at this point we should have read
	 * the magic number, so we're _not_ a the beginning of the file
	 */
	while(!feof(in)) {
		if(!remains) {
			DPRINT("no space left, reallocating\n");
			llist=realloc(llist,(num+RBI)*sizeof(lesson_t));
			if(!llist) {
				ERROR("can't allocate memory (%s)\n", strerror(errno));
				exit(1);
			}
			remains=RBI;
		}

		if(fread(&llist[num],1,sizeof(lesson_t),in)==sizeof(lesson_t)) {
			/*
			 * Hey ! we successfully read a lesson_t !
			 * 
			 * Champagne !
			 */
			DDO(lesson_show(llist[num]));
			num++;
			remains--;
		} else {
			DPRINT("couldn't read a complete lesson_t\n");
		}
	}
	fclose(in);

	DPRINT("EOF, read %d lesson_t_n, now adjusting memory\n",num);
	llist=realloc(llist,num*sizeof(lesson_t));
	if(!llist) {
		ERROR("can't allocate memory (%s)\n", strerror(errno));
		exit(1);
	}

	btree=btree_make_equilibrated(num,llist);
	
	free(llist);
	
	return btree;
}

lesson_t btree_get_smallest_and_remove(bt_t **btree)
{
	bt_t *tree;
	lesson_t l;

	memset(&l,0,sizeof(lesson_t));

	if(!btree)
		return l;

	tree=*btree;

	if(!tree)
		return l;

	if(tree->earlier) {
		return btree_get_smallest_and_remove(&tree->earlier);
	} else {
		*btree=tree->later;
		l=tree->lesson;
		free(tree);
		return l;
	}	
}

lesson_t btree_get_smallest(bt_t *btree)
{
	lesson_t l;

	memset(&l,0,sizeof(lesson_t));

	if(!btree)
		return l;

	if(btree->earlier) {
		return btree_get_smallest(btree->earlier);
	} else {
		btree=btree->later;
		l=btree->lesson;
		return l;
	}	
}

void btree_show(bt_t *btree)
{
	/*
	 * This is a dirty debug fonction and you will understand it alone
	 *
	 * OK, it's not so hard...
	 */
	static int inc_level=0;
	int i;
	

	if(!inc_level)
		printf("ROOT\n");

	inc_level++;

	for(i=0;i<inc_level-1;i++)
		printf(" |   ");
	printf(" +-> ");

	if(!btree) {
		printf("(nil)\n");
		inc_level--;
		return;
	}
	lesson_show(btree->lesson);

	btree_show(btree->earlier);
	btree_show(btree->later);

	inc_level--;

	if(!inc_level)
		printf("END\n");
}

lesson_t btree_find_next(bt_t *btree, lesson_t l)
{
	bt_t *cur=btree, *last=NULL;
	int min_diff=INT_MIN, temp;

	l.name[0]=0;

	DPRINT("seeking lesson ");
	DDO(lesson_show(l));

	while(cur) {
		temp=lesson_compare_end(l,cur->lesson);
		if(temp<0) {
			DPRINT("lesson occurs earlier (%d min) than ",temp);
			DDO(lesson_show(cur->lesson));
			
			if(min_diff<temp) {
				DPRINT("closer than the closest (%d<%d)\n",min_diff,temp);
				min_diff=temp;
				last=cur;
			}
			cur=cur->earlier;
			if(!cur)
				return last->lesson;

		} else {
			DPRINT("lesson occurs later (%d min) than ", temp);
			DDO(lesson_show(cur->lesson));
			
			//if(last)
			//	return last->lesson;
			//if(!cur->later)
			//	return cur->lesson;
			cur=cur->later;
			if (!cur)
				return last->lesson;
		}	
	}

	return l;
}
